@@ -1,22 +1,12 @@
+from __future__ import annotations
+
 import sys
 import warnings
+from collections.abc import Callable
 from enum import Enum
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Callable,
-    Dict,
-    Generic,
-    List,
-    Optional,
-    Tuple,
-    Type,
-    TypeVar,
-    Union,
-    overload,
-)
+from typing import TYPE_CHECKING, Any, Generic, TypeVar, Union, overload
 
-from pypika.terms import Term
+from pypika_tortoise.terms import Term
 
 from tortoise.exceptions import ConfigurationError, ValidationError
 from tortoise.validators import Validator
@@ -52,7 +42,7 @@ NO_ACTION = OnDelete.NO_ACTION
 
 class _FieldMeta(type):
     # TODO: Require functions to return field instances instead of this hack
-    def __new__(mcs, name: str, bases: Tuple[Type, ...], attrs: dict) -> type:
+    def __new__(mcs, name: str, bases: tuple[type, ...], attrs: dict) -> type:
         if len(bases) > 1 and bases[0] is Field:
             # Instantiate class with only the 1st base class (should be Field)
             cls = type.__new__(mcs, name, (bases[0],), attrs)
@@ -85,7 +75,7 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
     These attributes needs to be defined when defining an actual field type.
 
     .. attribute:: field_type
-        :annotation: Type[Any]
+        :annotation: type[Any]
 
         The Python type the field is.
         If adding a type as a mixin, _FieldMeta will automatically set this to that.
@@ -113,7 +103,7 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
         Is this field able to be DB-generated?
 
     .. attribute:: function_cast
-        :annotation: Optional[pypika.Term] = None
+        :annotation: Optional[pypika_tortoise.Term] = None
 
         A casting term that we need to apply in case the DB needs emulation help.
 
@@ -149,45 +139,43 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
     """
 
     # Field_type is a readonly property for the instance, it is set by _FieldMeta
-    field_type: Type[Any] = None  # type: ignore
+    field_type: type[Any] = None  # type: ignore
     indexable: bool = True
     has_db_field: bool = True
     skip_to_python_if_native: bool = False
     allows_generated: bool = False
-    function_cast: Optional[Callable[[Term], Term]] = None
+    function_cast: Callable[[Term], Term] | None = None
     SQL_TYPE: str = None  # type: ignore
     GENERATED_SQL: str = None  # type: ignore
 
     # These methods are just to make IDE/Linters happy:
     if TYPE_CHECKING:
 
-        def __new__(cls, *args: Any, **kwargs: Any) -> "Field[VALUE]":
+        def __new__(cls, *args: Any, **kwargs: Any) -> Field[VALUE]:
             return super().__new__(cls)
 
         @overload
-        def __get__(self, instance: None, owner: Type["Model"]) -> "Field[VALUE]": ...
+        def __get__(self, instance: None, owner: type[Model]) -> Field[VALUE]: ...
 
         @overload
-        def __get__(self, instance: "Model", owner: Type["Model"]) -> VALUE: ...
+        def __get__(self, instance: Model, owner: type[Model]) -> VALUE: ...
 
-        def __get__(
-            self, instance: Optional["Model"], owner: Type["Model"]
-        ) -> "Field[VALUE] | VALUE": ...
+        def __get__(self, instance: Model | None, owner: type[Model]) -> Field[VALUE] | VALUE: ...
 
-        def __set__(self, instance: "Model", value: VALUE) -> None: ...
+        def __set__(self, instance: Model, value: VALUE) -> None: ...
 
     def __init__(
         self,
-        source_field: Optional[str] = None,
+        source_field: str | None = None,
         generated: bool = False,
-        primary_key: Optional[bool] = None,
+        primary_key: bool | None = None,
         null: bool = False,
         default: Any = None,
         unique: bool = False,
-        db_index: Optional[bool] = None,
-        description: Optional[str] = None,
-        model: "Optional[Model]" = None,
-        validators: Optional[List[Union[Validator, Callable]]] = None,
+        db_index: bool | None = None,
+        description: str | None = None,
+        model: Model | None = None,
+        validators: list[Validator | Callable] | None = None,
         **kwargs: Any,
     ) -> None:
         if (index := kwargs.pop("index", None)) is not None:
@@ -237,13 +225,13 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
         self.index = bool(db_index)
         self.model_field_name = ""
         self.description = description
-        self.docstring: Optional[str] = None
-        self.validators: List[Union[Validator, Callable]] = validators or []
+        self.docstring: str | None = None
+        self.validators: list[Validator | Callable] = validators or []
         # TODO: consider making this not be set from constructor
-        self.model: Type["Model"] = model  # type: ignore
-        self.reference: "Optional[Field]" = None
+        self.model: type[Model] = model  # type: ignore
+        self.reference: Field | None = None
 
-    def to_db_value(self, value: Any, instance: "Union[Type[Model], Model]") -> Any:
+    def to_db_value(self, value: Any, instance: type[Model] | Model) -> Any:
         """
         Converts from the Python type to the DB type.
 
@@ -259,11 +247,6 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
         if value is not None and not isinstance(value, self.field_type):
             value = self.field_type(value)  # pylint: disable=E1102
 
-        if self.__class__ in self.model._meta.db.executor_class.TO_DB_OVERRIDE:
-            value = self.model._meta.db.executor_class.TO_DB_OVERRIDE[self.__class__](
-                self, value, instance
-            )
-
         self.validate(value)
         return value
 
@@ -311,22 +294,37 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
         """
         return {}
 
-    def _get_dialects(self) -> Dict[str, dict]:
+    def _get_dialects(self) -> dict[str, dict]:
         ret = {}
-        for dialect in [key for key in dir(self) if key.startswith("_db_")]:
-            item = {}
+        for dialect in dir(self):
+            if not dialect.startswith("_db_"):
+                continue
             cls = getattr(self, dialect)
+            d = cls.__dict__
             try:
-                cls = cls(self)
+                obj = cls(self)
             except TypeError:
                 pass
-            for key, val in cls.__dict__.items():
-                if not key.startswith("_"):
-                    item[key] = val
-            ret[dialect[4:]] = item
+            else:
+                props = {
+                    prop: getattr(obj, prop)
+                    for prop in dir(cls)
+                    if isinstance(getattr(cls, prop), property)
+                }
+                d = {**d, **props}
+
+            ret[dialect[4:]] = {k: v for k, v in d.items() if not k.startswith("_")}
+
         return ret
 
-    def get_db_field_types(self) -> Optional[Dict[str, str]]:
+    def get_db_field_type(self) -> str:
+        """
+        Returns the DB field type for this field for the current dialect.
+        """
+        dialect = self.model._meta.db.capabilities.dialect
+        return self.get_for_dialect(dialect, "SQL_TYPE")
+
+    def get_db_field_types(self) -> dict[str, str] | None:
         """
         Returns the DB types for this field.
 
@@ -335,12 +333,17 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
         """
         if not self.has_db_field:  # pragma: nocoverage
             return None
+        default = self.SQL_TYPE
         return {
-            "": getattr(self, "SQL_TYPE"),
+            "": default,
             **{
-                dialect: _db["SQL_TYPE"]
-                for dialect, _db in self._get_dialects().items()
-                if "SQL_TYPE" in _db
+                dialect: sql_type
+                for dialect, sql_type in (
+                    (key[4:], self.get_for_dialect(key[4:], "SQL_TYPE"))
+                    for key in dir(self)
+                    if key.startswith("_db_")
+                )
+                if sql_type != default
             },
         }
 
@@ -351,8 +354,19 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
         :param dialect: The requested SQL Dialect.
         :param key: The attribute/method name.
         """
-        dialect_data = self._get_dialects().get(dialect, {})
-        return dialect_data.get(key, getattr(self, key, None))
+        try:
+            dialect_cls = getattr(self, f"_db_{dialect}")  # throws AttributeError if not present
+            dialect_value = getattr(dialect_cls, key)  # throws AttributeError if not present
+        except AttributeError:
+            pass
+        else:  # we have dialect_cls and dialect_value, so lets use it
+            # it could be that dialect_value is a computed property, like in CharField._db_oracle.SQL_TYPE,
+            # and therefore one first needs to instantiate dialect_cls
+            if isinstance(dialect_value, property):
+                return getattr(dialect_cls(self), key)
+            return dialect_value
+        # If there is nothing special defined, return the value of self
+        return getattr(self, key, None)
 
     def describe(self, serializable: bool) -> dict:
         """
@@ -399,14 +413,14 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
                 }
         """
 
-        def _type_name(typ: Type) -> str:
+        def _type_name(typ: type) -> str:
             if typ.__module__ == "builtins":
                 return typ.__name__
             if typ.__module__ == "typing":
                 return str(typ).replace("typing.", "")
             return f"{typ.__module__}.{typ.__name__}"
 
-        def type_name(typ: Any) -> Union[str, List[str]]:
+        def type_name(typ: Any) -> str | list[str]:
             try:
                 return typ._meta.full_name
             except (AttributeError, TypeError):
@@ -419,7 +433,7 @@ class Field(Generic[VALUE], metaclass=_FieldMeta):
                 except TypeError:
                     return str(typ)
 
-        def default_name(default: Any) -> Optional[Union[int, float, str, bool]]:
+        def default_name(default: Any) -> int | float | str | bool | None:
             if isinstance(default, (int, float, str, bool, type(None))):
                 return default
             if callable(default):
